
#include <list>

#include "interfaces/SMPDBInterface.h"
#include "interfaces/abstract/all.h"
#include "interfaces/idapro/all.h"
#include "base/SMPDataFlowAnalysis.h"
#include "base/SMPBasicBlock.h"
#include "base/SMPFunction.h"
#include "base/SMPProgram.h"
#include "base/SMPStaticAnalyzer.h"

using namespace std;


// Used for binary search by function number in SMPStaticAnalyzer.cpp
//  to trigger debugging output and find which instruction in which
//  function is causing a crash.
bool SMPBinaryDebug = false;


static bool IsChunkUnshared(ea_t ChunkAddr, ea_t FuncHeadStart, ea_t FuncHeadEnd) 
{
        bool Unshared = true;
        SMP_xref_t CurrXrefs;

        // See if any xref reaches this ChunkAddr from outside the FuncHead range.
        for (bool ok = CurrXrefs.SMP_first_to(ChunkAddr, XREF_ALL);     ok;     ok = CurrXrefs.SMP_next_to()) {
                ea_t FromAddr = CurrXrefs.GetFrom();
                if ((FromAddr != 0) && (CurrXrefs.GetIscode())) {
                        // We found a code xref that comes to the ChunkAddr. Whether it is a fall-through or
                        //  a jump/call, it is unshared only if it comes from within the FuncHead address range.
                        if ((FromAddr < FuncHeadStart) || (FromAddr >= FuncHeadEnd)) {
                                Unshared = false;
                                break;
                        }
                }
        }

        return Unshared;
} // end of SMPProgram::IsChunkUnshared()



void STARS_IDA_Function_t::MarkSharedChunks()
{

	char name[1000];
	ea_t FirstEA=this->get_startEA();
	::get_func_name(FirstEA,name,sizeof(name));
        // Determine if we are dealing with shared chunks.
        ea_t FuncHeadLastAddr = 0;
        size_t ChunkCounter = 0;
        //func_tail_iterator_t FuncTail((func_t*)*dynamic_cast<STARS_IDA_Function_t*>(this->GetFuncInfo()));
        func_tail_iterator_t FuncTail(the_func);// (func_t*)*(dynamic_cast<STARS_IDA_Function_t*>(this->GetFuncInfo())));
        for (bool ChunkOK = FuncTail.main(); ChunkOK; ChunkOK = FuncTail.next()) 
	{
                const area_t &CurrChunk = FuncTail.chunk();
                ++ChunkCounter;
                if (1 == ChunkCounter) { // head chunk
                        FuncHeadLastAddr = CurrChunk.endEA;
                }
                else { // a tail chunk
                        if (IsChunkUnshared(CurrChunk.startEA, FirstEA, FuncHeadLastAddr)) 
			{
//                                this->UnsharedChunks = true;
//                                SMP_msg("INFO: Interface Found unshared tail chunk for %s at %lx\n",
//                                        name, (unsigned long) CurrChunk.startEA);
                        }
                        else 
			{
                                this->SharedChunks = true;
//                                SMP_msg("INFO: Interface Found tail chunk for %s at %lx\n",
//                                        name, (unsigned long) CurrChunk.startEA);
                        }
                }
        }

}

bool STARS_IDA_Function_t::IsMultiEntry()
{
	ea_t FirstEA=this->get_startEA();
        func_tail_iterator_t FuncTail(the_func);// (func_t*)*(dynamic_cast<STARS_IDA_Function_t*>(this->GetFuncInfo())));
        size_t CallTargetCount = 0; // how many addresses in this function are called?

        if (this->HasSharedChunks() || this->UnsharedChunks) {
                for (bool ChunkOK = FuncTail.main(); ChunkOK; ChunkOK = FuncTail.next()) {
                        const area_t &CurrChunk = FuncTail.chunk();
                        ea_t addr = CurrChunk.startEA; // Start with just the beginning addrs of chunks.
                        // Determine whether the instruction is a call target by looking
                        //  at its cross references and seeing if it has "TO" code xrefs.
                        SMP_xref_t xrefs, Distant_xrefs;
                        for (bool ok = xrefs.SMP_first_to(addr, XREF_FAR); ok; ok = xrefs.SMP_next_to()) {
                                ea_t DistantAddr = xrefs.GetFrom();
                                if ((DistantAddr != 0) && (xrefs.GetIscode())) {
                                        // Now we see if the distant instruction is in another function.
                                        STARS_Function_t *SourceFunc = SMP_get_func(DistantAddr);
                                        if (NULL != SourceFunc) {
                                                if (SourceFunc->get_startEA() != FirstEA) {
                                                        ++CallTargetCount;
                                                        break;
                                                }
                                        }
                                }
                        } // end for all xrefs
                } // end for all chunks
        }

        return  CallTargetCount > 1;
} 


void STARS_IDA_Function_t::UpdateXrefs()
{

	size_t ChunkCounter=0;
	func_tail_iterator_t FuncTail(the_func);
	enum cref_t NearJump = (cref_t)(fl_JN | XREF_USER);
	enum cref_t FarJump = (cref_t)(fl_JF | XREF_USER);

 	for (bool ChunkOK = FuncTail.main(); ChunkOK; ChunkOK = FuncTail.next()) {
		const area_t &CurrChunk = FuncTail.chunk();

		// Find the instructions for each chunk, audit the xrefs.
		for (ea_t addr = CurrChunk.startEA; addr < CurrChunk.endEA; addr = SMP_get_item_end(addr)) {
			flags_t InstrFlags = SMP_getFlags(addr);
			if (SMP_isHead(InstrFlags) && SMP_isCode(InstrFlags)) {
				// Fill cmd structure with disassembly of instr
				insn_t LocalCmd;
				ulong LocalFeatures;
				if (!SMPGetCmd(addr, LocalCmd, LocalFeatures)) {
					SMP_msg("ERROR: SMPGetCmd failed from MDAuditJumpXrefs at %lx\n",
						(unsigned long) addr);
				}
				else {
					// Determine whether the instruction is a jump target by looking
					//  at its cross references and seeing if it has "TO" code xrefs.
					SMP_xref_t xrefs, Distant_xrefs;
					for (bool ok = xrefs.SMP_first_to(addr, XREF_FAR); ok; ok = xrefs.SMP_next_to()) {
						ea_t DistantAddr = xrefs.GetFrom();
						if ((DistantAddr != 0) && (xrefs.GetIscode())) {
							// Now we see if the distant instruction has an xref to this instruction.
							bool FoundDistantXref = false;
							for (bool ok2 = Distant_xrefs.SMP_first_from(DistantAddr, XREF_FAR); ok2; ok2 = Distant_xrefs.SMP_next_from()) {
								ea_t TargetAddr = Distant_xrefs.GetTo();
								if (TargetAddr == addr) {
									FoundDistantXref = true;
									break;
								}
							}
							if (!FoundDistantXref) {
								SMP_msg("WARNING: Missing code Xref from %lx to %lx\n",
									(unsigned long) DistantAddr, (unsigned long) addr);
								long SignedAddrDiff = (long) (DistantAddr - addr);
								if ((SignedAddrDiff < -128) || (SignedAddrDiff > 127)) {
									add_cref(DistantAddr, addr, FarJump);
								}
								else {
									add_cref(DistantAddr, addr, NearJump);
								}
							}
						}
					} // end for all "to" xrefs
					// Now check the "from" xrefs to see if the target inst has the corresponding "to" xref.
					for (bool ok = xrefs.SMP_first_from(addr, XREF_FAR); ok; ok = xrefs.SMP_next_from()) {
						ea_t DistantAddr = xrefs.GetTo();
						if ((DistantAddr != 0) && (xrefs.GetIscode())) {
							// Now we see if the distant instruction has an xref to this instruction.
							bool FoundDistantXref = false;
							for (bool ok2 = Distant_xrefs.SMP_first_to(DistantAddr, XREF_FAR); ok2; ok2 = Distant_xrefs.SMP_next_to()) {
								ea_t SourceAddr = Distant_xrefs.GetFrom();
								if (SourceAddr == addr) {
									FoundDistantXref = true;
									break;
								}
							}
							if (!FoundDistantXref) {
								SMP_msg("WARNING: Missing code Xref to %lx from %lx\n",
									(unsigned long) DistantAddr, (unsigned long) addr);
								long SignedAddrDiff = (long) (DistantAddr - addr);
								if ((SignedAddrDiff < -128) || (SignedAddrDiff > 127)) {
									add_cref(DistantAddr, addr, FarJump);
								}
								else {
									add_cref(DistantAddr, addr, NearJump);
								}
							}
						}
					} // end for all "from" xrefs
				} // end if (!SMPGetCmd() ... else ...
			} // end if (IsHead() and IsCode())
		} // end for all addrs in chunk
	} // end for all chunks

	return;
}

void STARS_IDA_Function_t::BuildFuncIR(SMPFunction *func)
{
	bool FoundAllCallers = false;

	list<SMPInstr *>::iterator FirstInBlock = func->Instrs.end();
		// For starting a basic block
	list<SMPInstr *>::iterator LastInBlock = func->Instrs.end();
		// Terminating a basic block

        set<ea_t> FragmentWorkList;  // Distant code fragments that belong to this function and need processing
        ea_t InstAddr; // grab address to help in debugging, conditional breakpoints, etc.
        ea_t PreviousIndirJumpAddr = BADADDR;
        enum cref_t NearJump = (cref_t)(fl_JN | XREF_USER);
        enum cref_t FarJump = (cref_t)(fl_JF | XREF_USER);



	func_tail_iterator_t FuncTail(the_func);

	// Cycle through all chunks that belong to the function.
	size_t ChunkCounter = 0;
	bool GoodRTL;
	func->BuiltRTLs = true;
 	for (bool ChunkOK = FuncTail.main(); ChunkOK; ChunkOK = FuncTail.next()) {
		const area_t &CurrChunk = FuncTail.chunk();
		++ChunkCounter;
#if 0
		if (CurrChunk.startEA < func->FirstEA) {
			func->FirstEA = CurrChunk.startEA;
		}
#endif
#if STARS_DEBUG_MEMORY_CORRUPTION
	bool DebugFlag = (0 == strcmp("sub_8063BE0", func->GetFuncName()));
#endif
		// Build the instruction and block lists for the function.
		for (ea_t addr = CurrChunk.startEA; addr < CurrChunk.endEA;
			addr = SMP_get_item_end(addr)) {
			flags_t InstrFlags = SMP_getFlags(addr);
			if (!SMP_isHead(InstrFlags)) {
				continue;
			}
			if (!SMP_isCode(InstrFlags)) { //data
				// Check for code xrefs from the data.
				SMP_xref_t xrefs;
				for (bool ok = xrefs.SMP_first_from(addr, XREF_ALL); ok; ok = xrefs.SMP_next_from()) {
					if ((xrefs.GetTo() != 0) && (!xrefs.GetIscode())) {
						// Found a code target, with its address in xrefs.to
						PrintDataToCodeXref(addr, xrefs.GetTo(), 0);
					}
				}
			}
			else { // code
				SMPInstr *CurrInst = new SMPInstr(addr);
				// Fill in the instruction data members.
#if SMP_DEBUG_CONTROLFLOW
				SMP_msg("SMPFunction::Analyze: calling CurrInst::Analyze.\n");
#endif
				CurrInst->Analyze();
				if (SMPBinaryDebug) {
					SMP_msg("Disasm:  %s \n", CurrInst->GetDisasm());
				}
#if SMP_COUNT_MEMORY_ALLOCATIONS
				SMPInstBytes += sizeof(*CurrInst);
#endif

#if SMP_USE_SSA_FNOP_MARKER
				if (func->Instrs.empty()) {
					// First instruction in function. We want to create a pseudo-instruction
					//  at the top of the function that can hold SSA DEFs for LiveIn names
					//  to the function. We use a floating point no-op as the pseudo-inst.
					//  The code address is one less than the start address of the function.
					SMPInstr *MarkerInst = new SMPInstr(STARS_SSA_MARKER_PSEUDO_ID);
					MarkerInst->AnalyzeMarker();
					GoodRTL = MarkerInst->BuildRTL();
					func->BuiltRTLs = (func->BuiltRTLs && GoodRTL);
					if (GoodRTL) {
						MarkerInst->SetGoodRTL();
					}
					assert(FirstInBlock == func->Instrs.end());
					func->Instrs.push_back(MarkerInst);
#if SMP_COUNT_MEMORY_ALLOCATIONS
					SMPInstBytes += sizeof(*MarkerInst);
#endif
				}
#endif

				// Find all functions that call the current function.
				SMP_xref_t CurrXrefs;
				if (!FoundAllCallers) {
					for (bool ok = CurrXrefs.SMP_first_to(addr, XREF_ALL); ok; ok = CurrXrefs.SMP_next_to()) {
						ea_t FromAddr = CurrXrefs.GetFrom();
						if (FromAddr != 0) {
							if (CurrXrefs.GetIscode()) {
								// Make sure it is not a fall-through. Must be a
								//  control-flow instruction of some sort, including
								//  direct or indirect calls or tail calls.
								SMPInstr CallInst(FromAddr);
								CallInst.Analyze();
								SMPitype CallType = CallInst.GetDataFlowType();
								if ((COND_BRANCH <= CallType) && (RETURN >= CallType)) {
									// Found a caller, with its call address in CurrXrefs.from
									func->AddCallSource(FromAddr);
								}
							}
							else { // Data xref
								PrintDataToCodeXref(FromAddr, addr, 0);
								func->PossibleIndirectCallTarget = true;
								// Code xrefs all precede the first data xref, so
								//  we do not need to process any more xrefs. One
								//  indirect data reference is enough to change the
								//  status of the function; additional references are
								//  not important.
								break;
							}
						}
					}
					FoundAllCallers = true; // only do this for first inst
				}

				SMPitype DataFlowType = CurrInst->GetDataFlowType();
				if ((DataFlowType == INDIR_CALL) || (DataFlowType == CALL)) {
					// See if IDA has determined the target of the call.
#if 0				// Phase ordering: Do this later.
					CurrInst->AnalyzeCallInst(func->FirstEA, func->FuncInfo->get_endEA());
#endif
					ea_t TargetAddr = CurrInst->GetCallTarget();
					bool LinkedToTarget = (BADADDR != TargetAddr);
					if (LinkedToTarget) {
						if (0 == TargetAddr) {
							SMP_msg("WARNING: Ignoring NULL call target (unreachable) at %lx\n", 
								(unsigned long) CurrInst->GetAddr());
						}
						else {
							if (INDIR_CALL == DataFlowType) {
								pair<set<ea_t>::iterator, bool> InsertResult = func->IndirectCallTargets.insert(TargetAddr);
								if (InsertResult.second) {
									func->AllCallTargets.push_back(TargetAddr);
								}
							}
							else {
								(void) func->AddDirectCallTarget(TargetAddr);
							}
						}
					}
					if (DataFlowType == INDIR_CALL) {
						func->IndirectCalls = true;
						func->UnresolvedIndirectCalls = (!LinkedToTarget);
					}
				} // end if INDIR_CALL or CALL
				else if (DataFlowType == INDIR_JUMP) {
					func->IndirectJumps = true;
#if STARS_AUDIT_INDIR_JUMP_XREFS
					PreviousIndirJumpAddr = addr;
#endif
				}
				else if (DataFlowType == RETURN) {
					func->HasReturnInst = true;
				}
				// Add call targets for tail call jumps.
				else if (CurrInst->IsBranchToFarChunk()) {
					ea_t FarTargetAddr = CurrInst->GetFarBranchTarget();
					if (BADADDR != FarTargetAddr) {
						assert((RETURN == DataFlowType) || (JUMP == DataFlowType) || (COND_BRANCH == DataFlowType));
						// Optimized tail calls, where the stack frame is down to zero at the call point,
						//  get RETURN as their DataFlowType. Might have to revisit that idea at some point. !!!!****!!!!
						if (func->FindDistantCodeFragment(FarTargetAddr)) {
							if (func->GetProg()->InsertUnsharedFragment(FarTargetAddr)) {
								// Fragment address was inserted in SMPProgram set, was not already there.
								pair<set<ea_t>::iterator, bool> InsertResult;
								InsertResult = FragmentWorkList.insert(FarTargetAddr);
								if (InsertResult.second) {
									SMP_msg("INFO: Found distant code fragment at %lx that can be added to func, reached from %lx\n",
										(unsigned long) FarTargetAddr, (unsigned long) addr);
#if 0
									if (FarTargetAddr < func->FirstEA) {
										func->FirstEA = FarTargetAddr;
									}
#endif
								}
								else {
									// These kind of fragments are generally only jumped to from one place,
									//  and jump back into the function that jumped into them. Very suspicious
									//  to encounter such a fragment more than once, and even if it happens,
									//  the insertion into the SMPProgram set should have failed due to already
									//  being present. This message and assertion should never be reached.
									SMP_msg("FATAL ERROR: Distant fragment at %lx reached from %lx already reached from same function.\n",
										(unsigned long) FarTargetAddr, (unsigned long) addr);
									assert(InsertResult.second); // sanity lost; shut down
								}
							}
							else { // Fragment address was already in SMPProgram set
								; // Probably added in loop at beginning that found unshared fragments.
#if 0
								// These kind of fragments are generally only jumped to from one place,
								//  and jump back into the function that jumped into them. Very suspicious
								//  to encounter such a fragment more than once.
								SMP_msg("WARNING: Distant fragment at %x reached from %x has already been processed.\n",
									FarTargetAddr, addr);
#endif
							}
						}
						else if (!func->GetProg()->IsUnsharedFragment(FarTargetAddr)) {
							(void) func->AddDirectCallTarget(FarTargetAddr);
						}
					}
				}

#if STARS_AUDIT_INDIR_JUMP_XREFS
				if (FirstInBlock == func->Instrs.end()) { // CurrInst will start a block
					if (CurrInst->HasNoCodeXrefs() && (BADADDR != PreviousIndirJumpAddr) && (addr != PreviousIndirJumpAddr)) {
						// This block appears unreachable, but it can probably be reached by
						//  the most recent indirect jump. IDA Pro sometimes thinks it has
						//  resolved an indirect jump completely but has only done so partially.
						SMP_msg("WARNING: Adding possible missing indirect jump code Xref to %lx from %lx\n",
							(unsigned long) addr, (unsigned long) PreviousIndirJumpAddr);
						long SignedAddrDiff = (long) (addr - PreviousIndirJumpAddr);
						if ((SignedAddrDiff < -128) || (SignedAddrDiff > 127)) {
							SMP_add_cref(PreviousIndirJumpAddr, addr, FarJump);
						}
						else {
							SMP_add_cref(PreviousIndirJumpAddr, addr, NearJump);
						}
						CurrInst->SetJumpTarget();
					}
				}
#endif

				// Before we insert the instruction into the instruction
				//  list, determine if it is a jump target that does not
				//  follow a basic block terminator. This is the special case
				//  of a CASE in a SWITCH that falls through into another
				//  CASE, for example. The first sequence of statements
				//  was not terminated by a C "break;" statement, so it
				//  looks like straight line code, but there is an entry
				//  point at the beginning of the second CASE sequence and
				//  we have to split basic blocks at the entry point.
				if ((FirstInBlock != func->Instrs.end()) && CurrInst->IsJumpTarget()) {
#if SMP_DEBUG_CONTROLFLOW
					SMP_msg("SMPFunction::Analyze: hit special jump target case.\n");
#endif
					LastInBlock = --(func->Instrs.end());
					SMPBasicBlock *NewBlock = new SMPBasicBlock(func, FirstInBlock,	LastInBlock);
					// If not the first chunk in the function, it is a shared
					//  tail chunk.
					if (ChunkCounter > 1) {
						NewBlock->SetShared();
					}
					FirstInBlock = func->Instrs.end();
					LastInBlock = func->Instrs.end();
					func->Blocks.push_back(NewBlock);
					func->BlockCount += 1;
				}

				// Build tree RTLs for the instruction.
				GoodRTL = CurrInst->BuildRTL();
				func->BuiltRTLs = (func->BuiltRTLs && GoodRTL);
				if (GoodRTL) {
					CurrInst->SetGoodRTL();
				}
#if SMP_DEBUG_BUILD_RTL
				else {
					SMP_msg("ERROR: Cannot build RTL at %lx for %s\n", 
						(unsigned long) CurrInst->GetAddr(), CurrInst->GetDisasm());
				}
#endif

#if SMP_DEBUG_CONTROLFLOW
		SMP_msg("SMPFunction::Analyze: putting CurrInst on list.\n");
#endif
				// Insert instruction at end of list.
				func->Instrs.push_back(CurrInst);

				// Find basic block leaders and terminators.
				if (FirstInBlock == func->Instrs.end()) {
#if SMP_DEBUG_CONTROLFLOW
					SMP_msg("SMPFunction::Analyze: setting FirstInBlock.\n");
#endif
#if SMP_USE_SSA_FNOP_MARKER
					if (2 == func->Instrs.size()) {
						// Just pushed first real instruction, after the fnop marker.
						FirstInBlock = func->Instrs.begin();
					}
					else {
						FirstInBlock = --(func->Instrs.end());
					}
#else
					FirstInBlock = --(func->Instrs.end());
#endif
				}
				if (CurrInst->IsBasicBlockTerminator()) {
#if SMP_DEBUG_CONTROLFLOW
		SMP_msg("SMPFunction::Analyze: found block terminator.\n");
#endif
					LastInBlock = --(func->Instrs.end());
					SMPBasicBlock *NewBlock = new SMPBasicBlock(func, FirstInBlock, LastInBlock);
					// If not the first chunk in the function, it is a shared
					//  tail chunk.
					if (ChunkCounter > 1) {
						NewBlock->SetShared();
					}
					FirstInBlock = func->Instrs.end();
					LastInBlock = func->Instrs.end();
					func->Blocks.push_back(NewBlock);
					func->BlockCount += 1;

				}
			} // end if (isCode(InstrFlags))
		} // end for (ea_t addr = CurrChunk.startEA; ... )

		// Handle the special case in which a function does not terminate
		//  with a return instruction or any other basic block terminator.
		//  Sometimes IDA Pro sees a call to a NORET function and decides
		//  to not include the dead code after it in the function. That
		//  dead code includes the return instruction, so the function no
		//  longer includes a return instruction and terminates with a CALL.
		if (FirstInBlock != func->Instrs.end()) {
			LastInBlock = --(func->Instrs.end());
			SMPBasicBlock *NewBlock = new SMPBasicBlock(func, FirstInBlock, LastInBlock);
			// If not the first chunk in the function, it is a shared
			//  tail chunk.
			if (ChunkCounter > 1) {
				NewBlock->SetShared();
			}
			FirstInBlock = func->Instrs.end();
			LastInBlock = func->Instrs.end();
			func->Blocks.push_back(NewBlock);
			func->BlockCount += 1;
		}
	} // end for (bool ChunkOK = ...)
}



bool STARS_IDA_Function_t::FindDistantCodeFragment(SMPFunction* func, STARS_ea_t TargetAddr) 
{
	ea_t FirstEA=this->get_startEA();
	bool PrivateFragment = false;
	bool AlreadyFound = func->GetProg()->IsUnsharedFragment(TargetAddr);
	STARS_Function_t *TargetFunc = SMP_get_func(TargetAddr);
	if ((!AlreadyFound) && TargetFunc) {
		// Determine if we are dealing with shared chunks.
		size_t ChunkCounter = 0;
		func_tail_iterator_t FuncTail((func_t*)*(dynamic_cast<STARS_IDA_Function_t*>(TargetFunc)));
		for (bool ChunkOK = FuncTail.main(); ChunkOK; ChunkOK = FuncTail.next()) {
			++ChunkCounter;
		}
		if (1 < ChunkCounter) {
			SMP_msg("INFO: Code fragment at %lx is shared chunk.\n", (unsigned long) TargetAddr);
		}
		else {
			bool JumpsBackIntoCurrentFunc = false;
			bool HasReturnInstruction = false;
			bool AllocatesStackFrame = false;
		 	for (bool ChunkOK = FuncTail.main(); ChunkOK; ChunkOK = FuncTail.next()) {
				const area_t &CurrChunk = FuncTail.chunk();
				++ChunkCounter;
				// Build the instruction and block lists for the function.
				for (ea_t addr = CurrChunk.startEA; addr < CurrChunk.endEA;	addr = SMP_get_item_end(addr)) {
					flags_t InstrFlags = SMP_getFlags(addr);
					if (SMP_isHead(InstrFlags) && SMP_isCode(InstrFlags)) {
						SMPInstr *CurrInst = new SMPInstr(addr);
						// Fill in the instruction data members.
						CurrInst->Analyze();
						// Search for two negative indicators (stack allocations and returns)
						//  and one positive indicator (jump back into this function).
						if (CurrInst->MDIsReturnInstr()) {
							HasReturnInstruction = true;
							break;
						}
						else if (CurrInst->MDIsFrameAllocInstr()) {
							AllocatesStackFrame = true;
							break;
						}
						else {
							SMPitype FlowType = CurrInst->GetDataFlowType();
							if ((JUMP == FlowType) || (INDIR_JUMP == FlowType)) {
								if (CurrInst->BuildRTL()) {
									ea_t FragmentJumpTarget = CurrInst->GetJumpTarget();
									if ((FragmentJumpTarget >= FirstEA) && (FragmentJumpTarget <= this->get_endEA())) {
										JumpsBackIntoCurrentFunc = true;
									}
								}
							}
						}
					} // end if isHead() and isCode()
				} // end for all addrs in chunk
			} // end for all chunks (there will only be one)
			PrivateFragment = (JumpsBackIntoCurrentFunc && (!HasReturnInstruction) && (!AllocatesStackFrame));
		} // end if (1 < ChunkCounter) ... else ...
	} // end if (!AlreadyFound && TargetFunc)
	return PrivateFragment;
} // end of STARSFunction::FindDistantCodeFragment()
